# -*- coding: utf-8 -*-

# Compatibility with Python 2.7
from __future__ import absolute_import

# Import necesary packages
import re
from warnings import warn
from functools import partial
from copy import copy, deepcopy
from operator import attrgetter
from collections import defaultdict
from six import iteritems, iterkeys, string_types, integer_types

# from cobra
from cobra.core.object import Object
from cobra.core.gene import Gene, ast2str, parse_gpr, eval_gpr
from cobra.util.context import resettable, get_context

# from mass
from mass.core import expressions
from mass.core.massmetabolite import MassMetabolite

# Class begins
## Set a float infinity (Compatibility with Python 2.7)
inf = float('inf')
## precompiled regular expressions
### Matches and/or in a gene reaction rule
and_or_search = re.compile(r'\(| and| or|\+|\)', re.IGNORECASE)
uppercase_AND = re.compile(r'\bAND\b')
uppercase_OR = re.compile(r'\bOR\b')
gpr_clean = re.compile(' {2,}')
### This regular expression finds any single letter compartment enclosed in
### square brackets at the beginning of the string.
### For example [c] : foo --> bar
compartment_finder = re.compile("^\s*(\[[A-Za-z]\])\s*:*")
### Regular expressions to match the arrows for building reactions from strings
_reversible_arrow_finder = re.compile("<(-+|=+)>")
_forward_arrow_finder = re.compile("(-+|=+)>")
_reverse_arrow_finder = re.compile("<(-+|=+)")

# Class definition
class MassReaction(Object):
	"""MassReaction is a class for holding kinetic information regarding a
	biochemical reaction in a mass.MassModel object

	Parameters
	----------
	id : string
		The identifier to associate with this reaction
	name : string
		A human readable name for the reaction
	subsystem : string
		The subsystem where the reaction is meant to occur
	reversible : bool
		The kinetic reversibility of the reaction

	Attributes
	----------
	ssflux : float or None
		The steady state flux for the reaction. Can store a steady state flux
		and be utilized in pseudo rate constnant calculations.
	"""
	def __init__(self, id=None, name="", subsystem=None, reversible=True,
				ssflux=None):
		"""Initialize the MassReaction Object"""
		# Check inputs to ensure they are the correct types
		if not isinstance(name, string_types):
			raise TypeError("name must be a string type")
		if not isinstance(subsystem, (string_types, type(None))):
			raise TypeError("subsystem must be a string type")
		if not isinstance(reversible, bool):
			raise TypeError("reversible must be a boolean")

		Object.__init__(self, id, name)
		self.subsystem = subsystem
		self._reversible = reversible
		self.ssflux = ssflux
		# The forward, reverse, and equilibrium constants as strings
		# for symbolic expressions
		self._sym_kf = ("kf_%s" % id)
		self._sym_kr = ("kr_%s" % id)
		self._sym_Keq = ("Keq_%s" % id)

		# The forward, reverse, and equilbrium constants for simulation
		self._forward_rate_constant = None
		# No reverse rate constant for an irreversible reaction.
		# Therefore initialized to 0.
		if self._reversible:
			self._reverse_rate_constant = None
			self._equilibrium_constant = None
		else:
			self._reverse_rate_constant = 0.
			self._equilibrium_constant = inf

		# The rate law equation for simulation
		self._rate_law = None
		self._rate_expr = None
		self._rtype = 1

		# A dictionary of metabolites and their stoichiometric
		# coefficients for this kinetic reaction
		self._metabolites = dict()

		# The compartments where partaking metabolites are located
		self._compartments = None

		# The massmodel that the reaction is associated with
		self._model = None

		# The genes associated with the kinetic reaction
		self._genes = set()
		self._gene_reaction_rule = ""

		# The Gibbs reaction energy assoicated with this reaction
		self._gibbs_reaction_energy = None

		# For cobra compatibility if desired and escher visualization
		self.objective_coefficient = 0.
		self.variable_kind = 'continuous'
		if self._reversible:
			self.lower_bound = -1000.
			self.upper_bound = 1000.
		else:
			self.lower_bound = 0.
			self.upper_bound = 1000.


	# Properties
	@property
	def reversible(self):
		"""Returns the kinetic reversible of the reaction"""
		return self._reversible

	@reversible.setter
	def reversible(self, value):
		"""Set the kinetic reversibility of the reaction. Will initialize to
		default values for the reverse rate constant when the kinetic
		reversibility is changed to True, and will set the reverse rate
		constant to 0 if changed to False

		Parameters
		----------
		reversible : bool
			True for kinetically reversible reaction, False for irreversible
		"""
		if not isinstance(value, bool):
			raise TypeError("Must be a boolean True or False")

		self._reversible = value
		if value:
			self._reverse_rate_constant = None
			self._equilibrium_constant = None
		else:
			self._reverse_rate_constant = 0.
			self._equilibrium_constant = inf

	@property
	def forward_rate_constant(self):
		"""Returns the forward rate constant associated with this reaction"""
		return self._forward_rate_constant

	@forward_rate_constant.setter
	def forward_rate_constant(self, value):
		"""Set the forward rate constant for this reaction"""
		self._forward_rate_constant = value

	@property
	def reverse_rate_constant(self):
		"""Returns the reverse rate constant associated with this reaction"""
		return self._reverse_rate_constant

	@reverse_rate_constant.setter
	def reverse_rate_constant(self, value):
		"""Set the reverse rate constant for this reaction.
		If the reaction is not reversible, warns the user"""
		if self._reversible:
			self._reverse_rate_constant = value
		else:
			warn("Cannot set reverse rate constant for irreversible reactions")

	@property
	def equilibrium_constant(self):
		"""Returns the equilibrium constant associated with this reaction"""
		return self._equilibrium_constant

	@equilibrium_constant.setter
	def equilibrium_constant(self, value):
		"""Set the equilibrium constant for this reaction"""
		if self._reversible:
			self._equilibrium_constant = value
		else:
			warn("Cannot set equilibrium constant for irreversible reactions")

	@property
	def rate_constants(self):
		"""Returns a list containing the rate constants."""
		return [self._forward_rate_constant, self._reverse_rate_constant]

	@property
	def parameters(self):
		"""Returns a dictionary of all reaction parameters"""
		param = dict()
		key_list = [self._sym_kf, self._sym_Keq,]
		attr_list = ["_forward_rate_constant", "_equilibrium_constant" ]
		if self.reversible:
			key_list += [self._sym_kr]
			attr_list += ["_reverse_rate_constant"]

		for i, attr in enumerate(attr_list):
			if self.__dict__[attr] is not None:
				param.update({key_list[i] : self.__dict__[attr]})
		return param

	@property
	def metabolites(self):
		"""Returns the metabolites of the reaction as a read-only copy"""
		return self._metabolites.copy()

	@property
	def reactants(self):
		"""Returns a list of the reactants for the reaction

		Identical to the method in cobra.core.reaction
		"""
		return [m for m, c in iteritems(self._metabolites) if c < 0]

	@property
	def products(self):
		"""Returns a list of the products for the reaction

		Identical to the method in cobra.core.reaction
		"""
		return [m for m, c in iteritems(self._metabolites) if c >= 0]

	@property
	def stoichiometry(self):
		"""Returns a list containing the stoichiometry of the reaction"""
		return [c for m, c in iteritems(self._metabolites)]

	@property
	def rate(self):
		"""Returns the rate law as a sympy expression"""
		if self._model is not None and self in self._model.custom_rates:
			return self._model.custom_rates[self]
		else:
			return expressions.generate_rate_law(self, rate_type=self._rtype,
									sympy_expr=True,  update_reaction=True)

	@property
	def model(self):
		"""Returns the massmodel the reaction is associated with"""
		return self._model

	@property
	def reaction(self):
		"""Return the reaction as a human readable string

		Similar to the method in cobra.core.reaction
		"""
		return self.build_reaction_string()

	@reaction.setter
	def reaction(self, reaction_string):
		"""Use a human readable string to set the reaction.
		The direction of the arrow is used to determine reversibility.

		Similar to the method in cobra.core.reaction
		Parameters
		----------
		reaction_string : string
			String representation of the reaction. For example:
			'A + B <=> C' for reversible reactions
			'A + B --> C' for irreversible reactions where A & B are reactants
			'A + B <-- C' for irreversible reactions where A & B are products

		Warnings
		--------
		Care must be taken when doing setting a reaction in this manner to
			ensure the reaction's id matches those in the model.

		For forward and reversible arrows, reactants are searched on the
			left side of the arrow while products are searched on the right.
			For reverse arrows, reactants are searched on the left side of
			the arrow while products are searched on the right.
		"""
		return self.build_reaction_from_string(reaction_string)

	@property
	def compartments(self):
		"""Returns a set of compartments that metabolites are in

		Identical to the method in cobra.core.reaction
		"""
		if self._compartments is None:
			self._compartments = {met.compartment for met in self._metabolites
								  if met.compartment is not None}
		return self._compartments

	@property
	def exchange(self):
		"""Whether or not this reaction is an exchange reaction
		Returns True if the reaction has either no products or reactants

		Identical to the boundary method in cobra.core.reaction

		.. note:: These are reactions with a sink or source
		"""
		return (len(self.metabolites) == 1 and
			not (self.reactants and self.products))

	@property
	def get_external_metabolite(self):
		"""Get an "external" metabolite for exchanges.	Primarily used for
		setting fixed concentrations in the MassModel.

		Returns a string representation of the external metabolite, or None if
		reaction is not an exchange.
		"""
		if self.exchange:
			for metab in self.metabolites:
				_c = re.search("^\w*\S(?!<\_)(\_\S+)$", metab.id)
				if _c is not None and not re.search("\_L$|\_D$", _c.group(1)):
					ext_metab = re.sub(_c.group(1), "_e", metab.id)
				else:
					ext_metab = "{}{}".format(metab.id, "_e")
			return ext_metab
		else:
			return None
	@property
	def genes(self):
		"""Returns a frozenset of the genes associated with the reaction"""
		return frozenset(self._genes)

	@property
	def gene_reaction_rule(self):
		"""Returns the gene reaction rule as a string"""
		return self._gene_reaction_rule

	@gene_reaction_rule.setter
	def gene_reaction_rule(self, new_rule):
		"""Set the gene reaction rule using a string, associated the new genes
		and dissociated the old genes for the reaction

		Similar to the method in cobra.core.reaction

		Parameters
		----------
		new_rule : string
			String representation of the new reaction rule
		"""
		if get_context(self):
			warn("Context management not implemented for gene reaction rules")

		self._gene_reaction_rule = new_rule.strip()
		try:
			_, gene_names = parse_gpr(self._gene_reaction_rule)
		except (SyntaxError, TypeError) as e:
			if "AND" in new_rule or "OR" in new_rule:
				warn("uppercase AND/OR found in rule '%s' for %s" %
						(new_rule, repr(self)))
				new_rule = uppercase_AND.sub("and", new_rule)
				new_rule = uppercase_OR.sub("or", new_rule)
				self.gene_reaction_rule = new_rule
				return
			warn("malformed gene_reaction_rule '%s' for %s" %
					(new_rule, repr(self)))
			tmp_str = and_or_search.sub('', self._gene_reaction_rule)
			gene_names = set((gpr_clean.sub(' ', tmp_str).split(' ')))
		if '' in gene_names:
			gene_names.remove('')
		old_genes = self._genes
		if self._model is None:
			self._genes = {Gene(i) for i in gene_names}
		else:
			massmodel_genes = self._model.genes
			self._genes = set()
			for id in gene_names:
				if massmodel_genes.has_id(id):
					self._genes.add(massmodel_genes.get_by_id(id))
				else:
					new_gene = Gene(id)
					# Must be new_gene._model due to inheritance
					# of cobra gene class
					new_gene._model = self._model
					self._genes.add(new_gene)
					massmodel_genes.append(new_gene)

		# Make the genes aware that it is involved in this reaction
		for g in self._genes:
			g._reaction.add(self)

		# Make the old genes aware that they are no
		# longer involved in this reaction
		for g in old_genes:
			if g not in self._genes: # if an old gene is not a new gene
				try:
					g._reaction.remove(self)
				except:
					warn("Could not remove old gene %s from reaction %s" %
						(g.id, self.id))

	@property
	def gene_name_reaction_rule(self):
		"""Display gene_reaction_rule with names

		Identical to the method in cobra.core.reaction

		Warnings
		--------
		Do NOT use this string for computation. It is intended to give a
			representation of the rule using more familiar gene names instead
			of the often cryptic ids.
		"""
		names = {i.id: i.name for i in self._genes}
		ast = parse_gpr(self._gene_reaction_rule)[0]
		return ast2str(ast, names=names)

	@property
	def functional(self):
		"""Check if all required enzymes for reaction are functional.

		Returns True if the gene-protein-reaction (GPR) rule is fulfilled for
			this reaction, or if reaction is not associated to a massmodel,
			otherwise returns False.

		Identical to the method in cobra.core.reaction
		"""
		if self._model:
			tree, _ = parse_gpr(self.gene_reaction_rule)
			return eval_gpr(tree, {gene.id for gene in self.genes if
								   not gene.functional})
		return True

	@property
	def gibbs_reaction_energy(self):
		"""Returns the Gibbs reaction energy of the reaction"""
		return self._gibbs_reaction_energy

	@gibbs_reaction_energy.setter
	def gibbs_reaction_energy(self, value):
		"""Set the Gibbs reaction energy of the reaction"""
		if not isinstance(value, (integer_types, float)):
			raise TypeError("Must be an integer or float")

		self._gibbs_reaction_energy = value

	# Shorthands
	@property
	def kf(self):
		"""Shorthand for getting the forward rate constant"""
		return self._forward_rate_constant

	@kf.setter
	def kf(self, value):
		"""Shorthand for setting the forward rate constant"""
		self.forward_rate_constant = value

	@property
	def kr(self):
		"""Shorthand for getting the reverse rate constant"""
		return self.reverse_rate_constant

	@kr.setter
	def kr(self, value):
		"""Shorthand for setting the reverse rate constant"""
		self.reverse_rate_constant = value

	@property
	def Keq(self):
		"""Shorthand for getting the equilibrium constant"""
		return self._equilibrium_constant

	@Keq.setter
	def Keq(self, value):
		"""Shorthand for setting the forward rate constant"""
		self.equilibrium_constant = value

	@property
	def S(self):
		"""Shorthand for the reaction stoichiometry"""
		return [c for m, c in iteritems(self._metabolites)]

	# Methods
	## Public
	def reverse_stoichiometry(self):
		"""Reverse the stoichiometry of the reaction, making the products
		into the reactants and the reactants into the products. Only works
		on reversible reactions.
		"""
		if not self.reversible:
			warn("Can only reverse stoichiometry for reversible reactions")
			return None
		for metab, coeff in iteritems(self.metabolites):
			self._metabolites[metab] = -1*coeff


	def generate_rate_law(self, rate_type=1, sympy_expr=False,
						update_reaction=False):
		"""Generates the rate law for the reaction as a human readable string.
		or as a sympy expression for simulation.

		The type determines which rate law format to return.
		For example: A <=> B

		type=1: kf*(A - B/Keq)
		type=2: kf*A - kr*B
		type=3: kr*(Keq*A - B)

		Parameters
		----------
		rate_type : int {1, 2, 3}, optional
			The type of rate law to display. Must be 1, 2, of 3.
			type 1 will utilize kf and Keq,
			type 2 will utilize kf and kr,
			type 3 will utilize kr and Keq.
		sympy_expr : bool, optional
			If True, will output a sympy expression, otherwise
			will output a human readable string.
		update_reaction : bool, optional
			If True, will update the MassReaction in addition to returning the
			rate law. Otherwise just return the rate law.

		Returns
		-------
		string representation or sympy expression of the rate law
		"""
		return expressions.generate_rate_law(self, rate_type, sympy_expr,
										update_reaction)

	def get_mass_action_ratio(self, sympy_expr=False):
		"""Get the mass action ratio for the reaction as
		a human readable string or a sympy expression for simulation.

		Parameters
		----------
		sympy_expr : bool, optional
			If True, will output a sympy expression, otherwise
			will output a human readable string.

		Returns
		-------
		string representation or sympy expression of the mass action ratio
		"""
		return expressions.get_mass_action_ratio(self, sympy_expr=sympy_expr)

	def get_disequilibrium_ratio(self, sympy_expr=False):
		"""Get the disequilibrium ratio for the reaction as
		a human readable string or a sympy expression for simulation.

		Parameters
		----------
		sympy_expr : bool, optional
			If True, will output a sympy expression, otherwise
			will output a human readable string.

		Returns
		-------
		string representation or sympy expression of the disequilibrium ratio
		"""
		return expressions.get_disequilibrium_ratio(self,sympy_expr=sympy_expr)

	def remove_from_model(self, remove_orphans=False):
		"""Removes the reaction from a massmodel.

		This removes all associations between a reaction, the associated
		massmodel, metabolites and genes.

		The change is reverted upon exit when using the massmodel as a context.

		Identical to the method in cobra.core.reaction

		Parameters
		----------
		remove_orphans : bool, optional
			Remove orphaned genes and metabolites from the massmodel as well
		"""
		return self._model.remove_reactions([self],remove_orphans)

	def copy(self):
		"""Copy a reaction.

		The rate constants, rate laws, referenced metabolites, and genes are
		also copied.
		"""
		# No references to massmodel when copying
		massmodel = self._model
		self._model = None
		for i in self._metabolites:
			i._model = None
		for i in self._genes:
			i._model = None

		# The reaction can be copied
		new_massreaction = deepcopy(self)
		# Restore the references
		self._model = massmodel
		for i in self._metabolites:
			i._model = massmodel
		for i in self._genes:
			i._model = massmodel

		return new_massreaction

	def get_coefficient(self, metabolite_id):
		"""Return the stoichiometric coefficients of a metabolite
		in the reaction.

		Similar to the method in cobra.core.reaction

		Parameters
		----------
		metabolite_id : string or mass.MassMetabolite object.
		"""
		if isinstance(metabolite_id, MassMetabolite):
			return self._metabolites[metabolite_id]

		_id_to_metabolites = {m.id: m for m in self._metabolites}
		return self._metabolites[_id_to_metabolites[metabolite_id]]

	def get_coefficients(self, metabolite_ids):
		"""Return the stoichiometric coefficients for a list of
		metabolites in the reaction.

		Identical to the method in cobra.core.reaction

		Parameters
		----------
		metabolite_ids : iterable
			Containing strings or mass.MassMetabolite objects.
		"""
		return map(self.get_coefficient, metabolite_ids)

	def add_metabolites(self, metabolites_to_add, combine=True,
						reversibly=True):
		"""Add metabolites and stoichiometric coefficients to the reaction.
		If the final coefficient for a metabolite is 0 then it is removed
		from the reaction.

		The change is reverted upon exit when using the massmodel as a context.

		Similar to the method in cobra.core.reaction

		Parameters
		----------
		metabolites_to_add : dict
			Dictionary with MassMetabolite objects or metabolite identifiers as
			keys and coefficients as values. If keys are strings (name of a
			metabolite) the reaction must already be part of a massmodel and a
			metabolite with the given name must exist in the massmodel.
		combine : bool, optional
			Describes behavior a metabolite already exists in the reaction.
			True causes the coefficients to be added.
			False causes the coefficient to be replaced.
		reversibly : bool, optional
			Whether to add the change to the context to make the change
			reversibly or not (primarily intended for internal use).
		"""
		for metabolite, coefficient in iteritems(metabolites_to_add):
			if not isinstance(metabolite, MassMetabolite):
				raise TypeError("%s is not a MassMetabolite object"
								% metabolite.id)

		old_coefficients = self.metabolites
		new_metabolites = []
		_id_to_metabolites = dict([(x.id, x) for x in self._metabolites])

		for metabolite, coefficient in iteritems(metabolites_to_add):
			met_id = str(metabolite)
			# If a metabolite already exists in the reaction then
			# just add them.
			if met_id in _id_to_metabolites:
				reaction_metabolite = _id_to_metabolites[met_id]
				if combine:
					self._metabolites[reaction_metabolite] += coefficient
				else:
					self._metabolites[reaction_metabolite] = coefficient
			else:
				# If the reaction is in a massmodel, ensure we aren't using
				# a duplicate metabolite.
				if self._model:
					try:
						metabolite = \
							self._model.metabolites.get_by_id(met_id)
					except KeyError as e:
						if isinstance(metabolite, MassMetabolite):
							new_metabolites.append(metabolite)
						else:
							# do we want to handle creation here?
							raise e
				elif isinstance(metabolite, string_types):
					raise ValueError("Reaction '%s' does not belong to a "
									 "massmodel. Either add the reaction to a "
									 "massmodel or use MassMetabolite objects "
									 "instead of strings as keys."
									 % self.id)
				self._metabolites[metabolite] = coefficient
				# make the metabolite aware it is involved in this reaction
				metabolite._reaction.add(self)

		for metabolite, the_coefficient in list(self._metabolites.items()):
			if the_coefficient == 0:
				# make the metabolite aware that it no longer participates
				# in this reaction
				metabolite._reaction.remove(self)
				self._metabolites.pop(metabolite)

		massmodel = self.model
		context = get_context(self)
		if context and reversibly:
			if combine:
				# Just subtract the metabolites that were added
				context(partial(
						self.subtract_metabolites, metabolites_to_add,
						combine=True, reversibly=False))
			else:
				# Reset the metabolites with add_metabolites
				mets_to_reset = {key: old_coefficients[
								massmodel.metabolites.get_by_any(key)[0]]
								for key in iterkeys(metabolites_to_add)}

	def subtract_metabolites(self, metabolites_to_subtract, combine=True,
							reversibly=True):
		"""This function will 'subtract' metabolites from a reaction, which
		means add the metabolites with -1*coefficient. If the final coefficient
		for a metabolite is 0 then the metabolite is removed from the reaction.

		The change is reverted upon exit when using the massmodel as a context.

		Similar to the method in cobra.core.reaction

		Parameters
		----------
		metabolites_to_subtract : dict
			Dictionary with MassMetabolite objects or metabolite identifiers as
			keys and coefficients as values. If keys are strings (name of a
			metabolite) the reaction must already be part of a massmodel and a
			metabolite with the given name must exist in the massmodel.
		combine : bool, optional
			Describes behavior a metabolite already exists in the reaction.
			True causes the coefficients to be added.
			False causes the coefficient to be replaced.
		reversibly : bool, optional
			Whether to add the change to the context to make the change
			reversibly or not (primarily intended for internal use).

		Note: A final coefficient < 0 implies a reactant.

		Warnings
		--------
		To add a cobra Metabolite object to a MassReaction object, the
			cobra Metabolite must first be converted to a MassMetabolite
			through the from_cobra method in the mass.core.massmetabolite class
		"""
		self.add_metabolites({
			k: -v for k, v in iteritems(metabolites_to_subtract)},
			combine=combine, reversibly=reversibly)

	def build_reaction_string(self, use_metabolite_names=False):
		"""Generate a human readable reaction string

		Parameters
		----------
		use_metabolite_names : bool, optional
			If True, use metabolite names instead of identifiers.
			Default is False.

		Similar to the method in cobra.core.reaction
		"""
		def format(number):
			return "" if number == 1 else str(number).rstrip(".") + " "

		id_type = "id"
		if use_metabolite_names:
			id_type = "name"
		reactant_bits = []
		product_bits = []
		for metab in sorted(self._metabolites, key=attrgetter("id")):
			coefficient = self._metabolites[metab]
			metab_name = str(getattr(metab, id_type))
			if coefficient >= 0:
				product_bits.append(format(coefficient) + metab_name)
			else:
				reactant_bits.append(format(abs(coefficient)) + metab_name)
		reaction_string = " + ".join(reactant_bits)
		if self._reversible:
			reaction_string += " <=> "
		else:
			reaction_string += " --> "
		reaction_string += " + ".join(product_bits)
		return reaction_string


	def check_mass_balance(self):
		"""Compute mass and charge balance for the reaction

		returns a dict of {element: amount} for unbalanced elements.
		"charge" is treated as an element in this dict
		This should be empty for balanced reactions.

		Identical to the method in cobra.core.reaction
		"""
		reaction_element_dict = defaultdict(int)
		for metabolite, coefficient in iteritems(self._metabolites):
			if metabolite.charge is not None:
				reaction_element_dict["charge"] += \
					coefficient * metabolite.charge
			if metabolite.elements is None:
				raise ValueError("No elements found in metabolite %s"
								 % metabolite.id)
			for element, amount in iteritems(metabolite.elements):
				reaction_element_dict[element] += coefficient * amount
		# filter out 0 values
		return {k: v for k, v in iteritems(reaction_element_dict) if v != 0}

	def get_compartments(self):
		"""Return a list of compartments the metabolites are in

		Identical to the method in cobra.core.reaction
		"""
		return list(self.compartments)

	def build_reaction_from_string(self, reaction_string, verbose=True,
								   fwd_arrow=None, rev_arrow=None,
								   reversible_arrow=None, term_split="+"):
		"""Builds reaction from reaction equation reaction_string using parser

		Takes a string and using the specifications supplied in the optional
		arguments infers a set of metabolites, metabolite compartments and
		stoichiometries for the reaction.  It also infers the reversibility
		of the reaction from the reaction arrow.

		Changes to the associated massmodel are reverted upon exit when using
		the massmodel as a context.

		Similar to the method in cobra.core.reaction.

		Parameters
		----------
		reaction_str : string
			A string containing a reaction formula (equation)
		verbose: bool, optional
			Setting verbosity of function
		fwd_arrow : re.compile, optional
			For forward irreversible reaction arrows
		rev_arrow : re.compile, optional
			For backward irreversible reaction arrows
		reversible_arrow : re.compile, optional
			For reversible reaction arrows
		term_split : string, optional
			Dividing individual metabolite entries. Default is '+'

		"""
		# Set the arrows
		forward_arrow_finder = _forward_arrow_finder if fwd_arrow is None \
			else re.compile(re.escape(fwd_arrow))
		reverse_arrow_finder = _reverse_arrow_finder if rev_arrow is None \
			else re.compile(re.escape(rev_arrow))
		reversible_arrow_finder = _reverse_arrow_finder \
			if reversible_arrow is None \
			else re.compile(re.escape(reversible_arrow))
		if self._model is None:
			warn("No massmodel found")
			massmodel = None
		else:
			massmodel = self._model
		found_compartments = compartment_finder.findall(reaction_string)
		if len(found_compartments) == 1:
			compartment = found_compartments[0]
		else:
			compartment = ""

		#Reversible reaction
		arrow_match = reversible_arrow_finder.search(reaction_string)
		if arrow_match is not None:
			self._reversible = True
			# Reactants left of the arrow, products on the right
			reactant_str = reaction_string[:arrow_match.start()].strip()
			product_str = reaction_string[arrow_match.end():].strip()
		else: # Irreversible reaction
			# Try forward reaction
			arrow_match = forward_arrow_finder.search(reaction_string)
			if arrow_match is not None:
				self._reversible = False
				# Reactants left of the arrow, products on the right
				reactant_str = reaction_string[:arrow_match.start()].strip()
				product_str = reaction_string[arrow_match.end():].strip()
			else: # Try the reverse arrow
				arrow_match = reverse_arrow_finder.search(reaction_string)
				if arrow_match is None:
					raise ValueError("No suitable arrow found in '%s'" %
										reaction_str)
				else:
					self._reversible = False
					# Reactants right of the arrow, products on the left
					reactant_str = reaction_string[arrow_match.end():].strip()
					product_str = reaction_string[:arrow_match.start()].strip()

		self.subtract_metabolites(self.metabolites, combine=True)

		for substr, factor in ((reactant_str, -1), (product_str, 1)):
			if len(substr) == 0:
				continue
			for term in substr.split(term_split):
				term = term.strip()
				if term.lower() == "nothing":
					continue
				if " " in term:
					num_str, met_id = term.split()
					num = float(num_str.lstrip("(").rstrip(")")) * factor
				else:
					met_id = term
					num = factor
				met_id += compartment
				try:
					met = massmodel.metabolites.get_by_id(met_id)
				except KeyError:
					if verbose:
						print("Unknown metabolite '%s' created" % met_id)
					met = MassMetabolite(met_id)
				self.add_metabolites({met: num})

	def knock_out(self):
		"""Knockout reaction by setting its rate_constants to 0.

		Similar to the method in cobra.core.reaction"""
		self.forward_rate_constant = 0.
		if self._reversible:
			self.reverse_rate_constant = 0.
			self.equilibrium_constant = inf

	## Internal
	def _associate_gene(self, cobra_gene):
		"""Associates a cobra.Gene object with a mass.MassReaction.

		Identical to the method in cobra.core.reaction

		Parameters
		----------
		cobra_gene : cobra.core.Gene.Gene
		"""
		self._genes.add(cobra_gene)
		cobra_gene._reaction.add(self)
		cobra_gene._model = self._model

	def _dissociate_gene(self, cobra_gene):
		"""Dissociates a cobra.Gene object with a mass.MassReaction.

		Identical to the method in cobra.core.reaction

		Parameters
		----------
		cobra_gene : cobra.core.Gene.Gene
		"""
		self._genes.discard(cobra_gene)
		cobra_gene._reaction.discard(self)

	def _set_id_with_model(self, value):
		"""Set the id of the MassReaction object to the associated massmodel.

		Similar to the method in cobra.core.reaction
		"""

		if value in self.massmodel.reactions:
			raise ValueError("The massmodel already contains a reaction with "
								"the id:", value)
		self._id = value
		self.massmodel.reactions._generate_index()

	def _update_awareness(self):
		"""Make sure all metabolites and genes that are associated with
		this reaction are aware of it.
		"""
		for metab in self._metabolites:
			metab._reaction.add(self)
		for gene in self._genes:
			gene._reaction.add(self)

	def _repr_html_(self):
		return """
			<table>
				<tr>
					<td><strong>Reaction identifier</strong></td>
					<td>{id}</td>
				</tr><tr>
					<td><strong>Name</strong></td>
					<td>{name}</td>
				</tr><tr>
					<td><strong>Memory address</strong></td>
					<td>{address}</td>
				</tr><tr>
					<td><strong>Subsystem</strong></td>
					<td>{subsystem}</td>
				</tr><tr>
					<td><strong>Stoichiometry</strong></td>
					<td>
						<p style='text-align:right'>{stoich_id}</p>
						<p style='text-align:right'>{stoich_name}</p>
					</td>
				</tr><tr>
					<td><strong>GPR</strong></td>
					<td>{gpr}</td>
				</tr><tr>
					<td><strong>Kinetic Reversibility</strong></td>
					<td>{reversibility}</td>
				</tr>
			</table>
		""".format(id=self.id, name=self.name,
				subsystem=self.subsystem, address='0x0%x' % id(self),
				stoich_id=self.build_reaction_string(),
				stoich_name=self.build_reaction_string(True),
				gpr=self.gene_reaction_rule,
				reversibility=self._reversible)


	# Module Dunders
	## All dunders are similar or identical to cobra.core.reaction dunders
	def __copy__(self):
		"""Create a copy of the mass reaction

		Similar to the method in cobra.core.reaction
		"""
		massreaction_copy = copy(super(MassReaction, self))
		return massreaction_copy

	def __deepcopy__(self, memo):
		"""Create a deepcopy of the mass reaction

		Similar to the method in cobra.core.reaction
		"""
		massreaction_deepcopy = deepcopy(super(MassReaction, self), memo)
		return massreaction_deepcopy

	def __setstate__(self, state):
		"""Probably not necessary to set _model as the mass.MassModel that
		contains self sets the _model attribute for all metabolites and
		genes in the reaction.

		However, to increase performance speed we do want to let the metabolite
		and gene know that they are employed in this reaction

		Similar to the method in cobra.core.reaction
		"""
		self.__dict__.update(state)
		for x in state["_metabolites"]:
			setattr(x, "_model", self._model)
			x._reaction.add(self)
		for x in state["_genes"]:
			setattr(x, "_model", self._model)
			x._reaction.add(self)

	def __add__(self, other):
		"""Add two mass reactions

		The stoichiometry will be the combined stoichiometry of the two
		reactions, and the gene reaction rule will be both rules combined by an
		and. All other attributes (i.e. rate constants) will match those of
		the first reaction, Return a new MassReaction object.

		Similar to the method in cobra.core.reaction
		"""
		new_massreaction = self.copy()
		new_massreaction += other
		return new_massreaction

	def __iadd__(self, other):
		"""Add two mass reactions

		The stoichiometry will be the combined stoichiometry of the two
		reactions, and the gene reaction rule will be both rules combined by an
		and. All other attributes (i.e. rate constants) will match those of the
		first reaction. Return the same MassReaction object with its updates

		Identical to the method in cobra.core.reaction
		"""
		self.add_metabolites(other._metabolites, combine=True)
		gpr1 = self.gene_reaction_rule.strip()
		gpr2 = other.gene_reaction_rule.strip()
		if gpr1 != "" and gpr2 != "":
			self.gene_reaction_rule = ("(%s) and (%s)" % \
										(self.gene_reaction_rule,
										other.gene_reaction_rule))
		elif gpr1 != "" and gpr2 == "":
			self.gene_reaction_rule = gpr1
		elif gpr1 == "" and gpr2 != "":
			self.gene_reaction_rule = gpr2
		return self

	def __sub__(self, other):
		"""Subtract two mass reactions

		The stoichiometry will be the combined stoichiometry of the two
		reactions, and the gene reaction rule will be both rules combined by an
		'and'. All other attributes (i.e. rate constants) will match those of
		the first reaction. Return a new MassReaction object.

		Similar to the method in cobra.core.reaction
		"""
		new_massreaction = self.copy()
		new_massreaction -= other
		return new_massreaction

	def __isub__(self, other):
		"""Subtract two mass reactions

		The stoichiometry will be the combined stoichiometry of the two
		reactions.

		Similar to the method in cobra.core.reaction
		"""
		self.subtract_metabolites(other._metabolites, combine = True)
		return self


	def __mul__(self, coefficient):
		"""Scale coefficients in a reaction by a given value.
		Return a new MassReaction object

		E.g. A -> B becomes 2A -> 2B.
		"""
		new_massreaction = self.copy()
		new_massreaction *= coefficient
		return new_massreaction

	def __imul__(self, coefficient):
		"""Scale coefficients in a reaction by a given value.
		Return the same MassReaction object with its updates

		E.g. A -> B becomes 2A -> 2B.

		Similar to the method in cobra.core.reaction
		"""
		self._metabolites = {met: value * coefficient
							for met, value in iteritems(self._metabolites)}
		return self

	def __str__(self):
		"""Create an id string with the stoichiometry

		Identical to the method in cobra.core.reaction
		"""
		return "{id}: {stoichiometry}".format(
			id=self.id, stoichiometry=self.build_reaction_string())
